"""Preprocess input files."""

from pathlib import Path

def concatenate(dest: Path, *sources: Path, basedir: Path) -> None:
    """Write contents of sources into dest.

    Each section is preceded by an HTML comment.
    """

    comment = """
<!--#slipbox-metadata
filename: {}
-->
"""
    with dest.open("w") as file:
        for src in sources:
            filename = str(src.relative_to(basedir))
            print(comment.format(filename), file=file)
            print(src.read_text(), file=file)



"""Utils."""

import contextlib
import os
from pathlib import Path
import shlex
import shutil
import subprocess
import sys
import tempfile
from typing import Any, Dict, Iterator, Optional

def pandoc() -> str:
    """Pandoc location."""
    return os.environ.get("PANDOC", "pandoc")

def check_requirements() -> bool:
    """Check if pandoc is installed."""
    return bool(shutil.which(pandoc()))

def sqlite_string(text: str) -> str:
    """Encode python string into sqlite string."""
    return "'{}'".format(text.replace("'", "''"))

@contextlib.contextmanager
def temporary_directory() -> Iterator[Path]:
    """Path to temporary directory."""
    with tempfile.TemporaryDirectory() as tempdir:
        yield Path(tempdir)

def run_command(cmd: str,
                variables: Optional[Dict[str, str]] = None,
                **kwargs: Any) -> int:
    """Run command with additional environment variables in variables.

    Output stdout and stderr, and return the error code.

    kwargs
    : Additional arguments to subprocess.run
    """
    env = os.environ.copy()
    if variables is not None:
        env.update(variables)
    proc = subprocess.run(shlex.split(cmd), env=env, check=False,
                          capture_output=True, **kwargs)
    if proc.stdout:
        print(proc.stdout.decode())
    if proc.stderr:
        print(proc.stderr.decode(), file=sys.stderr)
    return proc.returncode

def insert_file_script(*files: Path, basedir: Path) -> str:
    """Create SQL query string to insert into the Files table."""
    sql = "INSERT INTO Files (filename) VALUES ({})"
    filenames = (sqlite_string(str(p.relative_to(basedir))) for p in files)
    return sql.format("), (".join(filenames))



"""Process CSV data."""

import csv
from pathlib import Path
from sqlite3 import Connection, IntegrityError
import sys
from typing import Callable, Sequence, Type

def warning(message: str, *information: str) -> None:
    """Show warning message."""
    print(f"[WARNING] {message}", file=sys.stderr)
    for info in information:
        print(f"  {info}", file=sys.stderr)

def run_sql_on_csv(conn: Connection,
                   path: Path,
                   sql: str,
                   types: Sequence[Type],
                   callback: Callable = None) -> None:
    """Run SQl query on CSV data."""
    cur = conn.cursor()
    with open(path) as file:
        reader = csv.reader(file)
        for row in reader:
            args = [t(a) for t, a in zip(types, row)]
            try:
                cur.execute(sql, args)
            except IntegrityError:
                if callback:
                    callback(*args)

def process_files(conn: Connection, path: Path) -> None:
    """Process Files data in path."""
    sql = "INSERT OR IGNORE INTO Files (filename) VALUES (?)"
    run_sql_on_csv(conn, path, sql, (str,))

def process_notes(conn: Connection, path: Path) -> None:
    """Process Notes data in path."""
    def fix(nid: int, title: str, filename: str) -> None:
        cur = conn.cursor()
        cur.execute("SELECT title, filename FROM Notes WHERE id = ?", (nid,))
        existing = cur.fetchone()
        warning(
            f"Duplicate ID: {nid}.",
            f"Could not insert note {title!r}.",
            f"Note {existing[0]!r} already uses the ID.",
            f"See {filename!r} or {existing[1]!r}."
        )

    sql = "INSERT INTO Notes (id, title, filename) VALUES (?, ?, ?)"
    run_sql_on_csv(conn, path, sql, (int, str, str), fix)

def process_links(conn: Connection, path: Path) -> None:
    """Process Links data in path."""
    sql = "INSERT OR IGNORE INTO Links (src, dest, annotation) VALUES (?, ?, ?)"
    run_sql_on_csv(conn, path, sql, (int, int, str))

def process_bibliography(conn: Connection, path: Path) -> None:
    """Process Bibliography data in path."""
    sql = "INSERT OR IGNORE INTO Bibliography (key, text) VALUES (?, ?)"
    run_sql_on_csv(conn, path, sql, (str, str))

def process_citations(conn: Connection, path: Path) -> None:
    """Process Citations data in path."""
    sql = "INSERT OR IGNORE INTO Citations (note, reference) VALUES (?, ?)"
    run_sql_on_csv(conn, path, sql, (int, str))

def process_clusters(conn: Connection, path: Path) -> None:
    """Process Clusters data in path."""
    sql = "INSERT OR IGNORE INTO Clusters (tag, src, dest, destType) VALUES (?, ?, ?, ?)"
    types = (str, int, str, str)
    cur = conn.cursor()
    with open(path) as file:
        reader = csv.reader(file)
        for row in reader:
            args = [t(a) for t, a in zip(types, row)]
            if args[3] == 'N':
                args[2] = int(args[2])
            try:
                cur.execute(sql, args)
            except IntegrityError:
                pass

def process_csvs(conn: Connection, basedir: Path) -> None:
    """Process CSV data in basedir."""
    process_files(conn, basedir/"files.csv")
    process_notes(conn, basedir/"notes.csv")
    process_links(conn, basedir/"links.csv")
    process_clusters(conn, basedir/"clusters.csv")
    process_bibliography(conn, basedir/"bibliography.csv")
    process_citations(conn, basedir/"citations.csv")



"""slipbox CLI commands."""

from .initializer import DotSlipbox
from .slipbox import Slipbox

def show_info(dot: DotSlipbox, note_id: int) -> None:
    """Print metadata associated with note ID."""
    sql = "SELECT title, filename FROM Notes WHERE id = ?"
    with Slipbox(dot) as slipbox:
        cur = slipbox.conn.cursor()
        cur.execute(sql, (note_id,))
        note = cur.fetchone()
        if note is not None:
            print(note_id)
            print(note[0])
            print(note[1])

def main(dot: DotSlipbox) -> None:
    """Compile notes into static page."""
    with Slipbox(dot) as slipbox:
        slipbox.run()



# type: ignore
"""Test slipbox.py."""

from time import time, sleep

import pytest

from .initializer import DotSlipbox
from .slipbox import Slipbox
from .utils import check_requirements, insert_file_script

def test_find_new_notes(tmp_path, sbox):
    """find_new_notes must only return existing files that aren't yet in the
    database and match the input patterns (*.md by default).
    """
    present = tmp_path/"present.md"
    absent = tmp_path/"absent.md"
    directory = tmp_path/"directory"
    txt = tmp_path/"ignore.txt"
    present.touch()
    absent.touch()
    directory.mkdir()
    txt.touch()

    sbox.conn.executescript(insert_file_script(present, basedir=sbox.basedir))
    assert list(sbox.find_new_notes()) == [absent]

def test_added_notes_pattern(tmp_path, sbox):
    """Results in sbox.find_new_notes() must match the input pattern."""
    sbox.dot.patterns = ('*.md', '*.txt')
    directory = tmp_path/"directory"
    markdown = tmp_path/"input.md"
    txt = tmp_path/"input.txt"
    tex = tmp_path/"ignore.tex"
    directory.mkdir()
    markdown.touch()
    txt.touch()
    tex.touch()
    assert sorted(sbox.find_new_notes()) == [markdown, txt]

def test_added_notes_in_db(tmp_path, sbox):
    """Results in sbox.find_new_notes() must not already be in the database."""
    new = tmp_path/"new.md"
    skip = tmp_path/"skip.md"
    new.touch()
    skip.touch()
    sbox.conn.executescript(insert_file_script(skip, basedir=sbox.basedir))
    assert list(sbox.find_new_notes()) == [new]

def test_added_notes_recursive(tmp_path, sbox):
    """sbox.find_new_notes() must find files recursively."""
    directory = tmp_path/"directory"
    new = directory/"new.md"
    directory.mkdir()
    new.touch()
    assert list(sbox.find_new_notes()) == [new]

def test_slipbox_context_manager(tmp_path):
    """Test database timestamp."""
    dot = DotSlipbox(tmp_path)
    with Slipbox(dot) as slipbox:
        assert slipbox.timestamp == 0.0
        slipbox.timestamp = time()
    with Slipbox(dot) as slipbox:
        assert slipbox.timestamp != 0.0

@pytest.mark.skipif(True, reason="flaky")
def test_modified_notes(tmp_path, sbox):
    """sbox.find_new_notes() must find modified notes after they are purged."""
    modified = tmp_path/"modified.md"
    not_modified = tmp_path/"not_modified.md"
    added = tmp_path/"added.md"

    sbox.conn.executescript(insert_file_script(modified, not_modified, basedir=sbox.basedir))
    sbox.timestamp = time()
    sleep(1)

    added.touch()
    modified.touch()
    sbox.purge()

    assert list(sbox.find_new_notes()) == [added, modified]

def test_purge(sbox, files_abc):
    """Input files must be purged from the database."""
    a_md, b_md, c_md = files_abc
    sbox.conn.executescript(insert_file_script(a_md, b_md, c_md, basedir=sbox.basedir))
    sbox.timestamp = time()
    sleep(1)

    b_md.unlink()
    c_md.touch()
    sbox.purge()

    result = sbox.conn.execute("SELECT filename FROM Files")
    remaining = sorted(filename for filename, in result)
    assert len(remaining) == 1

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_run(files_abc, capsys, sbox):
    """There must be no suggestions when running for the first time."""
    file_a, file_b, file_c = files_abc
    file_a.write_text("# 0 A\n\nA.\n[B](#2 '/a').\n")
    file_b.write_text("# 1 B\n\nB.\n[C](#2).\n")
    file_c.write_text("# 2 C\n\nC.\n")

    slipbox = sbox
    slipbox.timestamp = time()

    slipbox.run()

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process(tmp_path, sbox):
    """Smoke test for slipbox.process."""
    input_file = tmp_path/"input.md"
    input_file.write_text("# 1 Test note\n\nHello, world!\n")
    assert not list(sbox.conn.execute("SELECT * FROM Html"))

    sbox.process([input_file])
    result = list(sbox.conn.execute("SELECT * FROM Sections"))
    # Check if the output HTML has sections.
    assert len(result) == 1

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_empty_file(tmp_path, sbox):
    """Scanned files that are empty shouldn't have entries in the database."""
    empty = tmp_path/"empty.md"
    empty.touch()
    sbox.process([empty])
    assert not list(sbox.conn.execute("SELECT * FROM Files"))
    assert not list(sbox.conn.execute("SELECT * FROM Notes"))
    assert not list(sbox.conn.execute("SELECT * FROM Tags"))
    assert not list(sbox.conn.execute("SELECT * FROM Links"))
    assert not list(sbox.conn.execute("SELECT * FROM Html"))
    assert not list(sbox.conn.execute("SELECT * FROM Sections"))
    assert not list(sbox.conn.execute("SELECT * FROM Bibliography"))
    assert not list(sbox.conn.execute("SELECT * FROM Citations"))

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_clusters(tmp_path, sbox):
    """Check if clusters are stored in db."""
    markdown = tmp_path/"test.md"
    markdown.write_text("# 0 Test\n\n#test/1.\n\n# 1 Dest\n\nTest.\n")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT tag, src, dest, destType FROM Clusters"))
    assert result == [('#test', 0, 1, 'N')]

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_clusters_nested(tmp_path, sbox):
    """Check if tag hierarchies get saved."""
    test_md = tmp_path/"test.md"
    test_md.write_text("# 0 Foo\n#foobar/1.\n\n# 1 Bar\nBar.\n\n# 2 Baz\n#foobarbaz/foobar.\n")
    sbox.process([test_md])
    result = sbox.conn.execute("SELECT tag, src, dest, destType FROM Clusters")
    assert sorted(result) == [('#foobar', 0, 1, 'N'), ('#foobarbaz', 2, '#foobar', 'T')]

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_filenames(tmp_path, sbox):
    """Filenames must be scanned correctly."""
    markdown = tmp_path/"foo.md"
    skip = tmp_path/"bar.md"
    markdown.write_text("# 0 Note\n\nBody.\n")
    skip.write_text("# 0 Note\n\nBody.\n")
    sbox.process([markdown])

    result = list(sbox.conn.execute("SELECT filename FROM Notes WHERE id = 0"))
    assert len(result) == 1
    assert markdown.samefile(sbox.basedir/result[0][0])

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_filenames0(tmp_path, sbox):
    """Filenames must be scanned correctly."""
    markdown = tmp_path/"bar.md"
    skip = tmp_path/"foo.md"
    markdown.write_text("# 0 Note\n\nBody.\n")
    skip.write_text("# 0 Note\n\nBody.\n")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT filename FROM Notes WHERE id = 0"))
    assert len(result) == 1
    assert markdown.samefile(sbox.basedir/result[0][0])

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_non_level1_headers(tmp_path, sbox):
    """Only level 1 headers must be considered as note headers."""
    markdown = tmp_path/"test.md"
    markdown.write_text("""# 0 Valid note header

Foo.

## 1 Invalid note header

Bar.
""")
    sbox.process([markdown])
    result = [nid for nid, in sbox.conn.execute("SELECT id FROM Notes")]
    assert len(result) == 1
    assert result == [0]

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_duplicate_existing_id(tmp_path, sbox, capsys):
    """slipbox.process must show a warning if a new note shares the ID of an
    existing note.

    The warning message must show the filenames of both notes.
    """
    file_a = tmp_path/"a.md"
    file_a.write_text("# 0 Existing note\n\nTest.\n")

    sbox.process([file_a])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes"))
    assert len(result) == 1
    assert result == [(0, "Existing note")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

    file_b = tmp_path/"b.md"
    file_b.write_text("# 0 Duplicate note\n\nTest.\n")

    sbox.process([file_b])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes"))
    assert len(result) == 1
    assert result == [(0, "Existing note")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr
    assert str(file_a.relative_to(sbox.basedir)) in stderr
    assert str(file_b.relative_to(sbox.basedir)) in stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_duplicate_ids_in_a_file(tmp_path, capsys, sbox):
    """If there are duplicate IDs in a file, only the first one must be saved.

    slipbox.process must show a warning when this happens.
    """
    markdown = tmp_path/"test.md"
    markdown.write_text("""# 0 First note

Foo.

# 0 Duplicate

Bar.
""")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes"))
    assert len(result) == 1
    assert result == [(0, "First note")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_empty_link_target(tmp_path, capsys, sbox):
    """slipbox.process must show a warning if there is a link with an empty target."""
    markdown = tmp_path/"test.md"
    markdown.write_text("# 0 Foo\n\n[Empty]().\n")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT * FROM Links"))
    assert not result

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_id_in_scientific_form(tmp_path, capsys, sbox):
    """Headers with non-integer IDs should be ignored."""
    markdown = tmp_path/"test.md"
    markdown.write_text("# 1e1 Invalid note ID\n\nTest.\n")
    sbox.process([markdown])

    result = list(sbox.conn.execute("SELECT * FROM Notes"))
    assert not result

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_non_text_titles(tmp_path, capsys, sbox):
    """Notes with non-text titles in the header must still be recognized."""
    file_a = tmp_path/"a.md"
    file_b = tmp_path/"b.md"
    file_c = tmp_path/"c.md"
    file_a.write_text("# 0 $1 + 1 = 2$\n\nTest.\n")
    file_b.write_text("# 1 [Note 0](#0)\n\nTest.\n")
    file_c.write_text("# 2 `print('Title')`\n\nTest.\n")

    sbox.process([file_a, file_b, file_c])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes ORDER BY id"))

    assert len(result) == 3
    assert result == [(0, "1 + 1 = 2"), (1, "Note 0"), (2, "print('Title')")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_tags_with_trailing_punctuation(tmp_path, capsys, sbox):
    """If a hashtag has trailing invalid symbols, only the prefix must be saved."""
    markdown = tmp_path/"test.md"
    markdown.write_text("# 0 Test\n\n#tag.\n#tags.\n#0.\n")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT tag FROM Tags ORDER BY tag"))
    assert result == [("#0",), ("#tag",), ("#tags",)]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_modify_non_notes(tmp_path, capsys, sbox):
    """Modification filters must be applied on non-notes too.

    But they shouldn't be saved in the database.
    Expect modifications:

    - Links with an empty target should be turned into plain text.
    - Links with an empty text should use the target as the text,
      surrounded by [brackets].
    - Hashtags should be turned into links.
    """
    markdown = tmp_path/"test.md"
    markdown.write_text("""# Not a note

[Foo]()

[](Bar)

#Baz

# 0 Note

Bye.
""")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT * FROM Notes"))
    assert result == [(0, 'Note', str(markdown.relative_to(sbox.basedir)))]

    html = list(sbox.conn.execute("SELECT body FROM Html"))
    assert len(html) == 1
    html = html[0][0]
    assert html

    assert "<p>Foo</p>" in html
    assert '<p> [<a href="Bar">Bar</a>]</p>' in html
    assert '<p><a href="##Baz">#Baz</a></p>' in html

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr



# type: ignore
"""Functions for mocking the database."""

from pathlib import Path
import sqlite3
from typing import Iterable, List

import pytest

from .initializer import initialize_database, DotSlipbox
from .slipbox import Slipbox

@pytest.fixture
def mock_db() -> Iterable[sqlite3.Connection]:
    """Create an empty mock database with all the necessary tables."""
    with sqlite3.connect(":memory:") as conn:
        initialize_database(conn)
        yield conn

@pytest.fixture
def sbox(tmp_path) -> Slipbox:
    """Create automatically configured Slipbox object."""
    dot = DotSlipbox(tmp_path)
    with Slipbox(dot) as slipbox:
        yield slipbox

@pytest.fixture
def files_abc(tmp_path) -> List[Path]:
    """Create files in tmp_path: a.md, b.md, c.md."""
    files = [tmp_path/"a.md", tmp_path/"b.md", tmp_path/"c.md"]
    for path in files:
        path.touch()
    yield files

@pytest.fixture
def mnote(tmp_path) -> Path:
    """Mock markdown note in tmp_path/test.md."""
    path = tmp_path/"test.md"
    path.write_text("# 0 Test\n\nTest note.\n")
    yield path



"""Sqlite database wrapper."""

from itertools import chain
from pathlib import Path
from time import time
from typing import Iterable, List, Tuple

from . import scan, page
from .initializer import DotSlipbox

class Slipbox:
    """Slipbox main functions."""
    def __init__(self, dot: DotSlipbox):
        self.conn = dot.database()
        self.dot = dot
        self.config = dot.config

    @property
    def basedir(self) -> Path:
        """Return base directory regardless of current working directory."""
        return self.dot.parent

    @property
    def timestamp(self) -> float:
        """Get timestamp from Meta table."""
        cur = self.conn.cursor()
        cur.execute("SELECT * FROM Meta WHERE key = 'timestamp'")
        _, value = cur.fetchone()
        return value

    @timestamp.setter
    def timestamp(self, value: float) -> None:
        """Set timestamp in Meta table."""
        self.conn.execute("UPDATE Meta SET value = ? WHERE key = 'timestamp'", (value,))
        self.conn.commit()

    def close(self) -> None:
        """Close database connection."""
        self.conn.close()

    def __enter__(self) -> "Slipbox":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None: # type: ignore
        self.close()

    def find_new_notes(self) -> Iterable[Path]:
        """Yield files that are not yet in the database."""
        patterns = self.dot.patterns
        for path in self.basedir.rglob('*'):
            if path.is_file() and scan.has_valid_pattern(path, patterns) \
                    and not scan.is_file_in_db(path.relative_to(self.basedir), self.conn):
                yield path

    def purge(self) -> Tuple[List[Path], List[Path]]:
        """Purge outdated/missing files from database."""
        modified = []
        deleted = []
        sql = "SELECT filename FROM Files"
        timestamp = self.timestamp
        cur = self.conn.cursor()
        cur.execute("PRAGMA foreign_keys=ON")
        for filename, in cur.execute(sql):
            path = self.basedir/filename
            if not path.exists():
                deleted.append(filename)
            elif scan.is_recently_modified(timestamp, path):
                modified.append(filename)
        cur.executemany("DELETE FROM Files WHERE filename IN (?)",
                        ((filename,) for filename in chain(modified, deleted)))
        self.conn.commit()
        return modified, deleted

    def process(self, paths: Iterable[Path]) -> None:
        """Process input files."""
        inputs = list(set(paths))
        for batch in scan.group_by_file_extension(inputs):
            scan.process_batch(self.conn, list(batch), self.config, self.basedir)
        self.timestamp = time()

    def compile(self) -> None:
        """Compile processed HTML into final output."""
        options = self.config.get("slipbox", "document_options")
        page.generate_complete_html(self.conn, options, self.basedir)

    def run(self) -> None:
        """Run all steps needed to compile output."""
        self.purge()
        notes = self.find_new_notes()
        self.process(notes)
        self.compile()



"""Build HTML from all notes in slipbox."""

from argparse import ArgumentParser
from pathlib import Path
import sys

from .app import main, show_info
from .initializer import DotSlipbox, default_config
from .utils import check_requirements

if __name__ == "__main__":
    if not check_requirements():
        sys.exit("[ERROR] pandoc not found.")

    parser = ArgumentParser(description="Generate a single-page HTML from your notes.")
    subparsers = parser.add_subparsers(dest="command")

    build = subparsers.add_parser("build", help="generate static site")

    info = subparsers.add_parser("info", help="show information about note")
    info.add_argument("id", type=int, help="note ID")

    defaults = default_config()
    init = subparsers.add_parser("init", help="initialize notes directory")
    init.add_argument("directory", nargs='?', type=Path, default=Path())
    init.add_argument("-c", "--content-options",
                      default=defaults.get("slipbox", "content_options"),
                      help="pandoc options for the content")
    init.add_argument("-d", "--document-options",
                      default=defaults.get("slipbox", "document_options"),
                      help="pandoc options for the output")
    init.add_argument("--convert-to-data-url", action="store_true",
                      default=defaults.getboolean("slipbox", "convert_to_data_url"),
                      help="convert local images links to data URL")

    args = parser.parse_args()

    dot_slipbox = DotSlipbox.locate(Path().resolve())

    command = args.command
    del args.command
    if command == "init":
        args.convert_to_data_url = str(args.convert_to_data_url)
        parent = args.directory
        del args.directory
        parent.mkdir(parents=True, exist_ok=True)
        DotSlipbox(parent, args)
        print(f"Initialized .slipbox in {parent.resolve()!s}.")
    elif dot_slipbox is not None:
        if command == "build":
            main(dot_slipbox)
        elif command == "info":
            show_info(dot_slipbox, args.id)
    else:
        sys.exit("could not find '.slipbox' in any parent directory.")



"""Initialize notes repository."""

from argparse import Namespace
from configparser import ConfigParser
from pathlib import Path
from shlex import quote
from sqlite3 import Connection, connect
import sys
from typing import List, Sequence, Optional

def initialize_database(conn: Connection) -> None:
    """Initialize database with schema.sql."""
    sql = Path(__file__).with_name("schema.sql").read_text()
    conn.executescript(sql)
    conn.commit()

def default_config() -> ConfigParser:
    """Create ConfigParser object with default options."""
    css = quote(str(Path(__file__).with_name("pandoc.css").resolve()))
    config = ConfigParser()
    config["slipbox"] = {
        "content_options": "--mathjax",
        "document_options": f"--mathjax -H {css}",
        "convert_to_data_url": "False",
    }
    return config

class DotSlipbox:
    """Initialized .slipbox/ directory."""
    def __init__(self, parent: Path, args: Optional[Namespace] = None, *, exit_: bool = True):
        """Initialize data.db, patterns and config.cfg in ./slipbox/."""
        self.parent = parent
        self.path = parent/".slipbox"
        self.existing = True
        if not self.path.exists():
            self.existing = False
            self.path.mkdir()
            database = self.path.joinpath("data.db")
            with connect(database) as conn:
                initialize_database(conn)
            config = default_config()
            if args is not None:
                config["slipbox"].update(vars(args))
            with open(self.path/"config.cfg", "w") as config_file:
                config.write(config_file)
            self.path.joinpath("patterns").write_text("*.md\n*.markdown\n")
        self.check_config(exit_)

    @property
    def patterns(self) -> List[str]:
        """Return list of glob patterns."""
        text = self.path.joinpath("patterns").read_text()
        return [pat for pat in text.split('\n') if pat]

    @patterns.setter
    def patterns(self, value: Sequence[str]) -> None:
        """Set glob patterns."""
        with self.path.joinpath("patterns").open("w") as file:
            for pattern in value:
                if pattern:
                    print(pattern, file=file)

    @property
    def config(self) -> ConfigParser:
        """Return ConfigParser object from .slipbox/config.cfg."""
        config = ConfigParser()
        config.read(self.path/"config.cfg")
        return config

    @staticmethod
    def locate(path: Path) -> Optional["DotSlipbox"]:
        """Find .slipbox in parent directories of path, or None.

        Input path must be absolute.
        """
        assert path.is_absolute()
        while not path.is_dir():
            path = path.parent
        parent = path.parent
        while parent != path:
            dot = path/".slipbox"
            if dot.is_dir():
                return DotSlipbox(path)
            path = parent
            parent = path.parent
        return None

    def database(self) -> Connection:
        """Create connection to .slipbox/data.db."""
        return connect(self.path/"data.db")

    def check_config(self, exit_: bool = True) -> bool:
        """Check .slipbox/config.cfg."""
        content_options = self.config.get("slipbox", "content_options")
        if "--strip-comments" in content_options:
            config_path = self.path.joinpath("config.cfg").resolve()
            if not self.existing:
                for path in self.path.iterdir():
                    path.unlink()
                self.path.rmdir()
            if exit_:
                sys.exit(f"invalid content_options value in {config_path!s}")
            return False
        return True



# type: ignore
"""Test initializer.py."""

from argparse import Namespace

from .initializer import DotSlipbox

def test_initialize(tmp_path):
    """Must initialize .slipbox in the parent directory."""
    dot = DotSlipbox(tmp_path).path
    assert dot.is_dir()
    assert dot.joinpath("data.db").is_file()
    assert dot.joinpath("config.cfg").is_file()
    assert dot.joinpath("patterns").is_file()

def test_dot_slipbox_locate(tmp_path):
    """find_dot_slipbox must look for .slipbox in parent directories."""
    assert DotSlipbox.locate(tmp_path) is None

    dot = DotSlipbox(tmp_path)
    assert DotSlipbox.locate(tmp_path).path == dot.path

    child = tmp_path/"child"
    child.mkdir()
    assert DotSlipbox.locate(child).path == dot.path

    for file in dot.path.iterdir():
        file.unlink()
    dot.path.rmdir()
    assert DotSlipbox.locate(child) is None
    assert DotSlipbox.locate(tmp_path) is None

def test_dot_slipbox_patterns(tmp_path):
    """Test patterns property."""
    dot = DotSlipbox(tmp_path)
    assert sorted(dot.patterns) == ['*.markdown', '*.md']

    patterns = ['*.rst', '*.txt']
    dot.patterns = patterns
    assert dot.patterns == patterns

def test_dot_slipbox_check_config_new(tmp_path):
    """.check_config must delete newly initialized data with invalid config."""
    args = Namespace(content_options="--strip-comments")
    DotSlipbox(tmp_path, args, exit_=False)
    assert not tmp_path.joinpath(".slipbox").exists()

def test_dot_slipbox_check_config_existing(tmp_path):
    """.check_config must not delete existing data with invalid config."""
    dot = DotSlipbox(tmp_path)
    assert dot.path.exists()
    assert dot.path.joinpath("config.cfg").is_file()
    assert dot.check_config()

    config = dot.config
    config["slipbox"]["content_options"] = "--strip-comments"
    with open(dot.path.joinpath("config.cfg"), "w") as config_file:
        config.write(config_file)

    dot = DotSlipbox(tmp_path, exit_=False)
    assert not dot.check_config(False)
    assert dot.path.joinpath("config.cfg").is_file()



#type: ignore
"""Test data.py."""

from . import data

def test_warning(capsys):
    """warning must print all messages to stderr."""
    data.warning("line1", "line2", "line3", "line4")
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr
    for i in range(1, 5):
        assert f"line{i}" in stderr



"""Look for files that must be compiled."""

from configparser import ConfigParser
from itertools import groupby
import fnmatch
import os
from pathlib import Path
import shlex
from sqlite3 import Connection
import sys
from typing import Iterable, List, Tuple, Union

from . import utils
from .data import process_csvs
from .preprocess import concatenate

def is_recently_modified(timestamp: float, filename: Union[Path, str]) -> bool:
    """Check if file has been modified after the timestamp."""
    return os.path.exists(filename) and os.path.getmtime(filename) >= timestamp

def is_file_in_db(path: Path, conn: Connection) -> bool:
    """Check if file is recorded in the database."""
    cur = conn.cursor()
    sql = "SELECT filename FROM Files WHERE filename = ?"
    for _ in cur.execute(sql, (str(path),)):
        return True
    return False

def has_valid_pattern(filename: Path, patterns: Iterable[str]) -> bool:
    """Check if filename matches one of the patterns."""
    for pattern in patterns:
        if fnmatch.fnmatch(str(filename), pattern):
            return True
    return False

def group_by_file_extension(files: Iterable[Path]) -> Iterable[Iterable[Path]]:
    """Generate an iterator for each file extension.

    Each file with no file extension is given its own iterator.
    """
    def key(filename: Union[str, Path]) -> Tuple[str, str]:
        root, ext = os.path.splitext(filename)
        return (ext, "") if ext else ("", root)
    groups = groupby(sorted(files, key=key), key=key)
    return map(lambda g: g[1], groups)

def build_command(input_: Path, output: str, options: str = "") -> str:
    """Construct a single pandoc command to run on inputs.

    inputs is a string of filenames separated by spaces.
    Each filename must be shlex.quoted if needed.
    Return an empty string if there are no input files.
    """
    assert input_.exists()
    data_dir = shlex.quote(str(Path(__file__).parent.resolve()))
    cmd = f"{utils.pandoc()} -Fpandoc-citeproc -Lzk.lua --section-divs " \
            f"--data-dir={data_dir} -Mlink-citations:true {options} " \
            "-o {} ".format(shlex.quote(output))
    return cmd + ' ' + str(input_)

def store_html_sections(conn: Connection, html: str, sources: List[Path], basedir: Path) -> None:
    """Insert Html and Sections entries for html and sources.

    html
    : HTML body text.
    sources
    : List of source files from which the html was generated.
    """
    if not html.strip() or not sources:
        return
    cur = conn.cursor()
    cur.execute("PRAGMA foreign_keys=ON")
    cur.execute("INSERT INTO Html(body) VALUES (?)", (html,))
    lastrowid = cur.lastrowid
    query = "SELECT DISTINCT id FROM Notes WHERE filename IN ({})".format(
        ", ".join(utils.sqlite_string(str(source.relative_to(basedir))) \
                  for source in sources))
    insert = "INSERT OR IGNORE INTO Sections(note, html) VALUES (?, ?)"
    cur2 = conn.cursor()
    for row in cur.execute(query):
        cur2.execute(insert, (row[0], lastrowid))

def process_batch(conn: Connection,
                  batch: List[Path],
                  config: ConfigParser,
                  basedir: Path) -> None:
    """Process batch of input notes."""
    convert_to_data_url = "1" \
        if config.getboolean("slipbox", "convert_to_data_url") \
        else ""
    with utils.temporary_directory() as tempdir:
        html = tempdir/"temp.html"
        preprocessed_input = tempdir/"input.md"
        concatenate(preprocessed_input, *batch, basedir=basedir)
        cmd = build_command(preprocessed_input, str(html),
                            config.get("slipbox", "content_options"))
        retcode = utils.run_command(cmd, dict(SLIPBOX_TMPDIR=str(tempdir),
                                              CONVERT_TO_DATA_URL=convert_to_data_url),
                                    cwd=basedir)
        if retcode:
            print("Scan failed.", file=sys.stderr)
            return
        process_csvs(conn, tempdir)
        store_html_sections(conn, html.read_text(), batch, basedir)
        conn.commit()



# type: ignore
"""Test app.py."""

import pytest

from .app import show_info
from .initializer import DotSlipbox
from .slipbox import Slipbox
from .utils import check_requirements

def test_show_info_timestamp_empty(tmp_path):
    """Calling show_info shouldn't affect the db timestamp."""
    dot = DotSlipbox(tmp_path)
    with Slipbox(dot) as slipbox:
        before = slipbox.timestamp
    show_info(dot, 0)
    with Slipbox(dot) as slipbox:
        after = slipbox.timestamp
    assert before == after

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_show_info_timestamp(tmp_path, mnote, sbox):
    """Calling show_info shouldn't affect the db timestamp."""
    sbox.process([mnote])
    dot = DotSlipbox(tmp_path)
    with Slipbox(dot) as slipbox:
        before = slipbox.timestamp
    show_info(dot, 0)
    with Slipbox(dot) as slipbox:
        after = slipbox.timestamp
    assert before == after

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_show_info_empty_stdout(tmp_path, capsys, mnote, sbox):
    """show_info shouldn't output anything if the note ID doesn't exist."""
    sbox.process([mnote])
    dot = DotSlipbox(tmp_path)
    show_info(dot, 1)
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_show_info_stdout(tmp_path, capsys, mnote, sbox):
    """show_info should output info of note with give note ID."""
    sbox.process([mnote])
    dot = DotSlipbox(tmp_path)
    show_info(dot, 0)
    stdout, stderr = capsys.readouterr()
    assert stdout
    assert not stderr



# type: ignore
"""Test scan.py."""

import os
import time

import pytest

from . import scan
from .utils import insert_file_script

def test_is_recently_modified(tmp_path):
    """is_recently_modified should return true iff the input file was modified
    after the specified timestamp.
    """
    before = time.time()
    path = tmp_path/"file"
    path.touch()
    os.utime(path, ns=(time.time_ns(), time.time_ns()))
    after = time.time()

    assert scan.is_recently_modified(before, path)
    assert not scan.is_recently_modified(after, path)

    os.utime(path, ns=(time.time_ns(), time.time_ns()))
    assert scan.is_recently_modified(after, path)

def test_is_file_in_db(mock_db, tmp_path):
    """Quick check for is_file_in_db."""
    conn = mock_db
    present = tmp_path/"present"
    absent = tmp_path/"absent"
    conn.executescript(insert_file_script(present, basedir=tmp_path))
    assert scan.is_file_in_db(present.name, conn)
    assert not scan.is_file_in_db(absent.name, conn)

def test_has_valid_pattern():
    """has_valid_pattern is true only for filenames that match at least one of
    the wildcard patterns."""
    patterns = ("*.md", "*.rst")
    assert scan.has_valid_pattern("a.md", patterns)
    assert scan.has_valid_pattern("a.rst", patterns)
    assert not scan.has_valid_pattern("a.tex", patterns)
    assert not scan.has_valid_pattern("a.md", (".md"))

def test_group_by_file_extension():
    """group_by_file_extension should split by file type.

    Files with no extension should be considered their own type.
    """
    files = ["a.md", "b.md", ".md", "c.tex", ".tex", ""]
    groups = list(map(list, scan.group_by_file_extension(files)))
    assert len(groups) == 5
    assert ["a.md", "b.md"] in groups
    assert [".md"] in groups
    assert ["c.tex"] in groups
    assert [".tex"] in groups
    assert [""] in groups

def test_group_by_file_extension_on_the_same_type():
    """group_by_file_extension should group files with the same type together."""
    files = [f"{c}.md" for c in "abcdefg"]
    groups = list(map(list, scan.group_by_file_extension(files)))
    assert len(groups) == 1
    assert files in groups

def test_build_command(tmp_path):
    """Sanity check for build_command."""
    input_file = tmp_path/"input.md"
    output = "output.html"
    options = "--mathjax"
    input_file.touch()

    cmd = scan.build_command(input_file, output, options)
    assert str(input_file) in cmd
    assert f"-o {output}" in cmd
    assert options in cmd

@pytest.mark.xfail
def test_build_command_when_input_file_does_not_exist(tmp_path):
    """build_command must fail if input file does not exist."""
    input_file = tmp_path/"input.md"
    scan.build_command(input_file, "output.html", "")



#type: ignore
"""Test preprocess.py."""
from .preprocess import concatenate

def test_concatenate_when_dest_exists(files_abc, tmp_path):
    """There must be an HTML comment between each file section in the
    destination file.

    Assume destination file exists.
    """
    dest, *sources = files_abc
    sources[0].write_text("B")
    sources[1].write_text("C")

    concatenate(dest, *sources, basedir=tmp_path)
    content = dest.read_text()
    template = """<!--#slipbox-metadata
filename: {}
-->"""

    assert template.format(str(sources[0].relative_to(tmp_path))) in content
    assert template.format(str(sources[1].relative_to(tmp_path))) in content

def test_concatenate_when_dest_does_not_exist(tmp_path):
    """preprocess.concatenate must work even if the destination file
    does not exist.
    """
    dest = tmp_path/"dest.md"
    source_a = tmp_path/"a.md"
    source_b = tmp_path/"b.md"
    source_a.write_text("A")
    source_b.write_text("B")
    assert not dest.exists()

    concatenate(dest, source_a, source_b, basedir=tmp_path)
    content = dest.read_text()
    template = """<!--#slipbox-metadata
filename: {}
-->"""

    assert template.format(str(source_a.relative_to(tmp_path))) in content
    assert template.format(str(source_b.relative_to(tmp_path))) in content

def test_concatenate_when_there_are_no_sources(tmp_path):
    """preprocess.concatenate must create an empty file."""
    dest = tmp_path/"dest.md"
    dest.write_text("Delete this.")
    concatenate(dest, basedir=tmp_path)

    content = dest.read_text()
    assert not content



"""Generate complete HTML containing all slipbox notes."""

from pathlib import Path
import shlex
from sqlite3 import Connection
import subprocess
from typing import Iterable

from .templates import Elem, render
from .utils import pandoc, temporary_directory

DUMMY_MARKDOWN = r"""::: {style="display:none"}
$\,$
```c
```
:::"""

def generate_active_htmls(conn: Connection) -> Iterable[str]:
    """Get HTML stored in the database for active sections."""
    sql = """
        SELECT body FROM Html WHERE id IN (SELECT html FROM Sections)
            ORDER BY id DESC
    """
    return (body for body, in conn.execute(sql))

def generate_data(conn: Connection) -> Iterable[str]:
    """Generate slipbox data in javascript."""
    for nid, title, filename in conn.execute("SELECT id, title, filename FROM Notes ORDER BY id"):
        yield f"window.query.db.add(new Model.Note({nid}, {title!r}, {filename!r}))"
    sql = "SELECT src, dest, annotation FROM ValidLinks"
    for src, dest, annotation in conn.execute(sql):
        yield f"""window.query.db.add(new Model.Link(
  window.query.note({src}), window.query.note({dest}), {annotation!r}))"""
    sql = "SELECT tag, src, dest, destType FROM Clusters ORDER BY tag, src, dest, destType"
    for tag, src, dest, dest_type in conn.execute(sql):
        if dest_type == "N":
            yield f"window.query.db.add(new Model.Cluster({tag!r}, {src}, {dest}, {dest_type!r}))"
        elif dest_type == "T":
            yield f"window.query.db.add(new Model.Cluster({tag!r}, {src}, {dest!r}, {dest_type!r}))"

def generate_javascript(conn: Connection) -> Iterable[str]:
    """Generate slipbox javascript code."""
    yield '<script type="module">'
    bundle = Path(__file__).parent/"data"/"frontend.js"
    yield bundle.read_text().strip()
    yield '</script>'
    yield '<script type="text/javascript">'
    yield "window.addEventListener('DOMContentLoaded', () => {"
    yield from generate_data(conn)
    yield "window.initSlipbox()"
    yield "})"
    yield "</script>"

def create_bibliography(conn: Connection) -> str:
    """Create bibliography HTML section from database entries."""
    sql = "SELECT key, text FROM Bibliography ORDER BY key"
    items = []
    for key, text in conn.execute(sql):
        items.append(Elem("dt", Elem("a", f"[@{key[4:]}]", href='#' + key)))
        items.append(Elem("dd", text))
    section = Elem("section",
                   Elem("h1", "References"),
                   Elem("dl", *items),
                   id="references",
                   title="References",
                   style="display:none",
                   **{"class": "level1"})
    return render(section)

def create_tags(conn: Connection) -> str:
    """Create HTML section that lists all tags."""
    rows = conn.execute("SELECT DISTINCT tag FROM Tags ORDER BY tag")
    tags = (row[0] for row in rows)
    items = (Elem("li", Elem("a", tag, href=f"#{tag}")) for tag in tags)
    section = Elem("section",
                   Elem("h1", "Tags"),
                   Elem("ul", *items),
                   id="tags",
                   title="Tags",
                   style="display:none",
                   **{"class": "level1"})
    return render(section)

def create_tag_page(conn: Connection, tag: str) -> str:
    """Create HTML section that lists all notes with the tag."""
    sql = """
        SELECT id, title FROM Tags NATURAL JOIN Notes WHERE tag = ? ORDER BY id
    """
    items = []
    for nid, title in conn.execute(sql, (tag,)):
        item = Elem("li", f"[{nid}] ", Elem("a", title, href=f"#{nid}"))
        items.append(item)
    section = Elem("section",
                   Elem("h1", Elem("a", tag, href="#tags", title="List of tags")),
                   Elem("ul", *items),
                   id=tag,
                   title=tag,
                   style="display:none",
                   **{"class": "level1"})
    return render(section)

def create_tag_pages(conn: Connection) -> str:
    """Create all tag pages."""
    rows = conn.execute("SELECT DISTINCT tag FROM Tags ORDER BY tag")
    tags = (row[0] for row in rows)
    return '\n'.join(create_tag_page(conn, tag) for tag in tags)

def create_reference_page(conn: Connection, reference: str) -> str:
    """Create HTML section that lists all notes that cite the reference."""
    sql = """
        SELECT note, title, text FROM Citations
            JOIN Notes ON Citations.note = Notes.id
                JOIN Bibliography ON Bibliography.key = Citations.reference
                    WHERE reference = ?
    """
    items = []
    text = ""
    for note, title, _text in conn.execute(sql, (reference,)):
        text = _text
        item = Elem("li", f"[{note}] ", Elem("a", title, href=f"#{note}"))
        items.append(item)
    section = Elem("section",
                   Elem("h1", Elem("a", '@' + reference[4:], href="#references")),
                   Elem("p", text),
                   Elem("ul", *items),
                   id=reference,
                   title=reference,
                   style="display:none",
                   **{"class": "level1"})
    return render(section)

def create_reference_pages(conn: Connection) -> str:
    """Create all reference pages."""
    rows = conn.execute("SELECT key FROM Bibliography ORDER BY key")
    references = (row[0] for row in rows)
    return '\n'.join(create_reference_page(conn, ref) for ref in references)

def generate_complete_html(conn: Connection, options: str, basedir: Path) -> None:
    """Create final HTML file with javascript."""
    with temporary_directory() as tempdir:
        script = tempdir/"script.js"
        html = tempdir/"cached.html"
        extra = tempdir/"extra.html"
        dummy = tempdir/"Slipbox.md"
        dummy.write_text(DUMMY_MARKDOWN)
        script.write_text('\n'.join(generate_javascript(conn)))
        html.write_text('\n'.join(generate_active_htmls(conn)))
        with open(extra, "w") as file:
            print(create_tag_pages(conn), file=file)
            print(create_tags(conn), file=file)
            print(create_reference_pages(conn), file=file)
            print(create_bibliography(conn), file=file)
        cmd = "{pandoc} {dummy} -H{script} {title} -B{html} -B{extra} --section-divs {opts}".format(
            pandoc=pandoc(), dummy=shlex.quote(str(dummy)), script=shlex.quote(str(script)),
            html=shlex.quote(str(html)), opts=options, extra=shlex.quote(str(extra)),
            title="--metadata title=Slipbox")
        subprocess.run(shlex.split(cmd), check=False, cwd=basedir)
